import itertools
import mmap
import os
import shutil
import tempfile
from collections.abc import Iterable as Itr
from pathlib import Path
from typing import (
    Union,
    Iterable,
    Iterator,
)

import numpy as np

from .document import DocumentArrayGetAttrMixin
from .search_ops import DocumentArraySearchOpsMixin
from .traversable import TraversableSequence
from ..document import Document

HEADER_NONE_ENTRY = (-1, -1, -1)
PAGE_SIZE = mmap.ALLOCATIONGRANULARITY


class DocumentArrayMemmap(
    TraversableSequence, DocumentArrayGetAttrMixin, DocumentArraySearchOpsMixin, Itr
):
    """
    Create a memory-map to an :class:`DocumentArray` stored in binary files on disk.

    Memory-mapped files are used for accessing :class:`Document` of large :class:`DocumentArray` on disk,
    without reading the entire file into memory.

    The :class:`DocumentArrayMemmap` on-disk storage consists of two files:
        - `header.bin`: stores id, offset, length and boundary info of each Document in `body.bin`;
        - `body.bin`: stores Documents continuously

    When loading :class:`DocumentArrayMemmap`, it only loads the content of `header.bin` into memory, while storing
    all `body.bin` data on disk. As `header.bin` is often much smaller than `body.bin`, memory is saved.

    This class is designed to work similarly as :class:`DocumentArray` but differs in the following aspects:
        - one can not set the attribute of elements in a :class:`DocumentArrayMemmap`;
        - one can not use slice to index elements in a :class:`DocumentArrayMemmap`;

    To convert between a :class:`DocumentArrayMemmap` and a :class:`DocumentArray`

    .. highlight:: python
    .. code-block:: python

        # convert from DocumentArrayMemmap to DocumentArray
        dam = DocumentArrayMemmap('./tmp')
        ...

        da = DocumentArray(dam)

        # convert from DocumentArray to DocumentArrayMemmap
        dam2 = DocumentArrayMemmap('./tmp')
        dam2.extend(da)
    """

    def __init__(self, path: str, key_length: int = 36):
        Path(path).mkdir(parents=True, exist_ok=True)
        self._header_path = os.path.join(path, 'header.bin')
        self._body_path = os.path.join(path, 'body.bin')
        self._key_length = key_length
        self._load_header_body()

    def reload(self):
        """Reload header of this object from the disk.

        This function is useful when another thread/process modify the on-disk storage and
        the change has not been reflected in this :class:`DocumentArray` object.

        This function only reloads the header, not the body.
        """
        self._load_header_body()

    def _load_header_body(self, mode: str = 'a'):
        if hasattr(self, '_header'):
            self._header.close()
        if hasattr(self, '_body'):
            self._body.close()

        open(self._header_path, mode).close()
        open(self._body_path, mode).close()

        self._header = open(self._header_path, 'r+b')
        self._body = open(self._body_path, 'r+b')

        tmp = np.frombuffer(
            self._header.read(),
            dtype=[
                ('', (np.str_, self._key_length)),  # key_length x 4 bytes
                ('', np.int64),  # 8 bytes
                ('', np.int64),  # 8 bytes
                ('', np.int64),  # 8 bytes
            ],
        )
        self._header_entry_size = 24 + 4 * self._key_length

        self._header_map = {
            r[0]: (idx, r[1], r[2], r[3])
            for idx, r in enumerate(tmp)
            if not np.array_equal((r[1], r[2], r[3]), HEADER_NONE_ENTRY)
        }
        self._body_fileno = self._body.fileno()
        self._start = 0
        if self._header_map:
            self._start = tmp[-1][1] + tmp[-1][3]
            self._body.seek(self._start)

    def __len__(self):
        return len(self._header_map)

    def extend(self, values: Iterable['Document']) -> None:
        """Extend the :class:`DocumentArrayMemmap` by appending all the items from the iterable.

        :param values: the iterable of Documents to extend this array with
        """
        for d in values:
            self.append(d, flush=False)
        self._header.flush()
        self._body.flush()

    def clear(self) -> None:
        """Clear the on-disk data of :class:`DocumentArrayMemmap`"""
        self._load_header_body('wb')

    def append(self, doc: 'Document', flush: bool = True) -> None:
        """
        Append :param:`doc` in :class:`DocumentArrayMemmap`.

        :param doc: The doc needs to be appended.
        :param flush: If set, then flush to disk on done.
        """
        value = doc.binary_str()
        l = len(value)  #: the length
        p = int(self._start / PAGE_SIZE) * PAGE_SIZE  #: offset of the page
        r = (
            self._start % PAGE_SIZE
        )  #: the remainder, i.e. the start position given the offset
        self._header.write(
            np.array(
                (doc.id, p, r, r + l),
                dtype=[
                    ('', (np.str_, self._key_length)),
                    ('', np.int64),
                    ('', np.int64),
                    ('', np.int64),
                ],
            ).tobytes()
        )
        self._header_map[doc.id] = (len(self._header_map), p, r, r + l)
        self._start = p + r + l
        self._body.write(value)
        if flush:
            self._header.flush()
            self._body.flush()

    def __getitem__(self, key: Union[int, str]) -> 'Document':
        if isinstance(key, str):
            pos_info = self._header_map[key]
            _, p, r, l = pos_info
            with mmap.mmap(self._body_fileno, offset=p, length=l) as m:
                return Document(m[r:])
        elif isinstance(key, int):
            return self[self._int2str_id(key)]
        else:
            raise TypeError(f'`key` must be int or str, but receiving {key!r}')

    def __delitem__(self, key: Union[int, str]):
        if isinstance(key, str):
            idx = self._str2int_id(key)
            str_key = key
        elif isinstance(key, int):
            idx = key
            str_key = self._int2str_id(idx)
        else:
            raise TypeError(f'`key` must be int or str, but receiving {key!r}')

        p = idx * self._header_entry_size
        self._header.seek(p, 0)
        self._header.write(
            np.array(
                (key, -1, -1, -1),
                dtype=[
                    ('', (np.str_, self._key_length)),
                    ('', np.int64),
                    ('', np.int64),
                    ('', np.int64),
                ],
            ).tobytes()
        )
        self._header.seek(0, 2)
        self._header.flush()
        self._header_map.pop(str_key)

    def _str2int_id(self, key: str) -> int:
        return self._header_map[key][0]

    def _int2str_id(self, key: int) -> str:
        p = key * self._header_entry_size
        self._header.seek(p, 0)
        d_id = np.frombuffer(
            self._header.read(4 * self._key_length), dtype=(np.str_, self._key_length)
        )
        self._header.seek(0, 2)
        return d_id[0]

    def __iter__(self) -> Iterator['Document']:
        for k in self._header_map.keys():
            yield self[k]

    def __setitem__(self, key: Union[int, str], value: 'Document') -> None:
        if isinstance(key, int):
            if 0 <= key < len(self):
                # override an existing entry
                self.append(value)
                self._header_map[self._int2str_id(key)] = self._header_map[value.id]
                del self[value.id]
            else:
                raise IndexError(f'`key`={key} is out of range')
        elif isinstance(key, str):
            value.id = key
            self.append(value)
        else:
            raise TypeError(f'`key` must be int or str, but receiving {key!r}')

    @classmethod
    def _flatten(cls, sequence):
        return itertools.chain.from_iterable(sequence)

    def __bool__(self):
        """To simulate ```l = []; if l: ...```

        :return: returns true if the length of the array is larger than 0
        """
        return len(self) > 0

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self._header_path == other._header_path
            and self._body_path == other._body_path
        )

    def __contains__(self, item: str):
        return item in self._header_map

    def prune(self) -> None:
        """Prune deleted Documents from this object, this yields a smaller on-disk storage. """
        tdir = tempfile.mkdtemp()
        dam = DocumentArrayMemmap(tdir, key_length=self._key_length)
        dam.extend(self)
        dam.reload()
        os.remove(self._body_path)
        os.remove(self._header_path)
        shutil.copy(os.path.join(tdir, 'header.bin'), self._header_path)
        shutil.copy(os.path.join(tdir, 'body.bin'), self._body_path)
        self.reload()



import os
import pickle
import shutil
from pathlib import Path
from datetime import datetime
from collections.abc import MutableMapping
from typing import Callable, Dict, Sequence, TYPE_CHECKING, Tuple, Union

from jina.logging.logger import JinaLogger
from .. import jinad_args, __root_workspace__
from ..models import DaemonID
from ..models.base import StoreItem, StoreStatus

if TYPE_CHECKING:
    from ..models.workspaces import WorkspaceItem
    from ..models.containers import ContainerItem


class BaseStore(MutableMapping):
    """The Base class for Daemon stores"""

    _kind = ''
    _status_model = StoreStatus

    def __init__(self):
        self._logger = JinaLogger(self.__class__.__name__, **vars(jinad_args))
        self.status = self.__class__._status_model()

    def add(self, *args, **kwargs) -> DaemonID:
        """Add a new element to the store. This method needs to be overridden by the subclass


        .. #noqa: DAR101"""
        raise NotImplementedError

    def update(self, *args, **kwargs) -> DaemonID:
        """Updates the element to the store. This method needs to be overridden by the subclass


        .. #noqa: DAR101"""
        raise NotImplementedError

    def delete(self, *args, **kwargs) -> DaemonID:
        """Deletes an element from the store. This method needs to be overridden by the subclass


        .. #noqa: DAR101"""
        raise NotImplementedError

    def clear(self) -> None:
        """Deletes an element from the store. This method needs to be overridden by the subclass


        .. #noqa: DAR101"""
        raise NotImplementedError

    def __iter__(self):
        return iter(self.status.items)

    def __len__(self):
        return len(self.status.items)

    def __repr__(self) -> str:
        return str(self.status.dict())

    def keys(self) -> Sequence['DaemonID']:
        """Get keys in the store

        :return: Keys in the local store
        """

        return self.status.items.keys()

    def values(self) -> Sequence[Union['WorkspaceItem', 'ContainerItem']]:
        """Get values in the store

        :return: Values in the local store
        """

        return self.status.items.values()

    def items(
        self,
    ) -> Sequence[Tuple['DaemonID', Union['WorkspaceItem', 'ContainerItem']]]:
        """Get items in the store

        :return: Items in the local store
        """

        return self.status.items.items()

    def __getitem__(self, key: DaemonID) -> Union['WorkspaceItem', 'ContainerItem']:
        """Fetch a Container/Workspace object from the store

        :param key: the key (DaemonID) of the object
        :return: the value of the object
        """
        return self.status.items[key]

    def __setitem__(self, key: DaemonID, value: StoreItem) -> None:
        """Add a Container/Workspace object to the store

        :param key: the key (DaemonID) of the object
        :param value: the value to be assigned
        """
        self.status.items[key] = value
        self.status.num_add += 1
        self.status.time_updated = datetime.now()

    def __delitem__(self, key: DaemonID) -> None:
        """Release a Container/Workspace object from the store

        :param key: the key (DaemonID) of the object


        .. #noqa: DAR201"""
        self.status.items.pop(key)
        self.status.num_del += 1
        self.status.time_updated = datetime.now()

    def __setstate__(self, state: Dict):
        self._logger = JinaLogger(self.__class__.__name__, **vars(jinad_args))
        now = datetime.now()
        self.status = self._status_model(**state)
        self.status.time_updated = now

    def __getstate__(self) -> Dict:
        return self.status.dict()

    @classmethod
    def dump(cls, func) -> Callable:
        """Dump store as a pickle to local workspace

        :param func: function to be wrapped
        :return: decorator for dump
        """

        def wrapper(self, *args, **kwargs):
            r = func(self, *args, **kwargs)
            filepath = os.path.join(__root_workspace__, f'{self._kind}.store')
            if Path(filepath).is_file():
                shutil.copyfile(filepath, f'{filepath}.backup')
            with open(filepath, 'wb') as f:
                pickle.dump(self, f)
            return r

        return wrapper

    @classmethod
    def load(cls) -> Union[Dict, 'BaseStore']:
        """Load store from a pickle in local workspace

        :return: Store from local or empty store
        """

        filepath = os.path.join(__root_workspace__, f'{cls._kind}.store')
        if Path(filepath).is_file() and os.path.getsize(filepath) > 0:
            with open(filepath, 'rb') as f:
                return pickle.load(f)
        else:
            return cls()

    def reset(self) -> None:
        """Calling :meth:`clear` and reset all stats """

        self.clear()
        self.status = self._status_model()



from queue import Empty
from threading import Thread
from typing import List, Optional

from fastapi import UploadFile

from jina.enums import RemoteWorkspaceState
from jina.helper import cached_property, colored
from jina.logging.logger import JinaLogger
from . import __task_queue__, daemon_logger, jinad_args
from .dockerize import Dockerizer
from .excepts import DockerImageException, DockerNetworkException
from .files import DaemonFile, workspace_files
from .helper import id_cleaner, get_workspace_path
from .models.id import DaemonID
from .models.workspaces import WorkspaceArguments, WorkspaceItem, WorkspaceMetadata
from .stores import workspace_store as store


class DaemonWorker(Thread):
    """Worker Thread for JinaD"""

    def __init__(
        self, id: 'DaemonID', files: List[UploadFile], name: str, *args, **kwargs
    ) -> None:
        super().__init__(name=f'{self.__class__.__name__}{name}', daemon=True)
        self.id = id
        self.files = files
        self._logger = JinaLogger(
            self.name, workspace_path=self.workdir, **vars(jinad_args)
        )
        self.start()

    @cached_property
    def arguments(self) -> WorkspaceArguments:
        """sets arguments in workspace store

        :return: pydantic model for workspace arguments
        """
        try:
            _args = store[self.id].arguments.copy(deep=True)
            _args.files.extend([f.filename for f in self.files] if self.files else [])
            _args.jinad.update(
                {
                    'build': self.daemon_file.build,
                    'dockerfile': self.daemon_file.dockerfile,
                }
            )
            _args.requirements = self.daemon_file.requirements
        except AttributeError:
            _args = WorkspaceArguments(
                files=[f.filename for f in self.files] if self.files else [],
                jinad={
                    'build': self.daemon_file.build,
                    'dockerfile': self.daemon_file.dockerfile,
                },
                requirements=self.daemon_file.requirements,
            )
        return _args

    @cached_property
    def metadata(self) -> WorkspaceMetadata:
        """sets metadata in workspace store

        :return: pydantic model for workspace metadata
        """
        image_id = self.generate_image()
        try:
            _metadata = store[self.id].metadata.copy(deep=True)
            _metadata.image_id = image_id
            _metadata.image_name = self.id.tag
        except AttributeError:
            _metadata = WorkspaceMetadata(
                image_id=image_id,
                image_name=self.id.tag,
                network=id_cleaner(self.network_id),
                workdir=self.workdir,
            )
        return _metadata

    @cached_property
    def workdir(self) -> str:
        """sets workdir for current worker thread

        :return: local directory where files would get stored
        """
        return get_workspace_path(self.id)

    @cached_property
    def daemon_file(self) -> DaemonFile:
        """set daemonfile for current worker thread

        :return: DaemonFile object representing current workspace
        """
        return DaemonFile(workdir=self.workdir, logger=self._logger)

    @cached_property
    def network_id(self) -> str:
        """create a docker network

        :return: network id
        """
        return Dockerizer.network(workspace_id=self.id)

    def generate_image(self):
        """build and create a docker image

        :return: image id
        """
        return Dockerizer.build(
            workspace_id=self.id, daemon_file=self.daemon_file, logger=self._logger
        )

    @cached_property
    def container_id(self) -> Optional[str]:
        """creates a container if run command is passed in .jinad file

        :return: container id, if created
        """
        if self.daemon_file.run:
            container, _, _ = Dockerizer.run_custom(
                workspace_id=self.id, daemon_file=self.daemon_file
            )
            return id_cleaner(container.id)
        else:
            return None

    def run(self) -> None:
        """
        Method representing the worker thread's activity
        DaemonWorker is a daemon thread responsible for the following tasks:
        During create:
        - store uploaded files in a local workspace
        - create a docker network for the workspace which would be used by all child containers
        - build a docker image to be used by all child containers
        - create a container if `run` command is passed
        During update:
        - update files in the local workspace
        - removes the workspace container, if any
        - recreate workspace container, if `run` command is passed
        """
        try:
            store.update(
                id=self.id,
                value=RemoteWorkspaceState.UPDATING
                if store[self.id].arguments
                else RemoteWorkspaceState.CREATING,
            )
            workspace_files(workspace_id=self.id, files=self.files, logger=self._logger)
            store.update(
                id=self.id,
                value=WorkspaceItem(
                    state=RemoteWorkspaceState.UPDATING,
                    metadata=self.metadata,
                    arguments=self.arguments,
                ),
            )

            # this needs to be done after the initial update, otherwise run won't find the necessary metadata
            # If a container exists already, kill it before running again
            previous_container = store[self.id].metadata.container_id
            if previous_container:
                self._logger.info(f'Deleting previous container {previous_container}')
                store[self.id].metadata.container_id = None
                del self.container_id
                Dockerizer.rm_container(previous_container)

            # Create a new container if necessary
            store[self.id].metadata.container_id = self.container_id
            store[self.id].state = RemoteWorkspaceState.ACTIVE

            self._logger.success(
                f'workspace {colored(str(self.id), "cyan")} is updated'
            )
        except DockerNetworkException as e:
            store.update(id=self.id, value=RemoteWorkspaceState.FAILED)
            self._logger.error(f'Error while creating the docker network: {e!r}')
        except DockerImageException as e:
            store.update(id=self.id, value=RemoteWorkspaceState.FAILED)
            self._logger.error(f'Error while building the docker image: {e!r}')
        except Exception as e:
            # TODO: how to communicate errors to users? users track it via logs?
            # TODO: Handle cleanup in case of exception
            store.update(id=self.id, value=RemoteWorkspaceState.FAILED)
            self._logger.error(f'{e!r}')


class ConsumerThread(Thread):
    """Consumer Thread for JinaD"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(daemon=True)
        # TODO: This is used for naming the worker, just appending doesn't make sense
        self._workers_count = 0

    def run(self) -> None:
        """
        Method representing the ConsumerThread's activity
        ConsumerThread is a daemon thread that waits for messages from the `__task_queue__`
        and starts a `DaemonWorker` for each message.
        """
        while True:
            try:
                workspace_id, files = __task_queue__.get()
                daemon_logger.info(
                    f'starting DaemonWorker{self._workers_count} for workspace {colored(workspace_id, "cyan")}'
                )
                DaemonWorker(
                    id=workspace_id, files=files, name=str(self._workers_count)
                )
                self._workers_count += 1
            except ValueError as e:
                daemon_logger.error(
                    f'got an invalid item in the queue. unable to process! {e!r}'
                )
            except Empty:
                pass



from collections.abc import MutableMapping
from typing import Optional, Union

import numpy as np

from ..score import NamedScore
from ...helper import typename
from ...proto import jina_pb2

__all__ = ['NamedScoreMapping']

if False:
    from google.protobuf.pyext._message import MessageMapContainer


class NamedScoreMapping(MutableMapping):
    """
    :class:`NamedScoreMapping` is one of the **primitive data type** in Jina.

    It offers a Pythonic interface to allow users access and manipulate
    :class:`jina.jina_pb2.NamedScoreMappingProto` object without working with Protobuf itself.

    It offers an interface to access and update scores as `NamedScore` as values of a `dict` with a string key.

    To create a :class:`NamedScoreMappingProto` object, simply:

        .. highlight:: python
        .. code-block:: python

            from jina.types.score.map import NamedScoreMapping
            scores = NamedScoreMapping()
            scores['score'] = 50

    :class:`NamedScoreMapping` can be built from ``jina_pb2.NamedScoreMappingProto`` (as a weak reference or a deep copy)

    :param scores: The scores to construct from, depending on the ``copy``,
        it builds a view or a copy from it.
    :type score: Optional[jina_pb2.NamedScoreMappingProto]
    :param copy: When ``scores`` is given as a :class:`NamedScoreMappingProto` object, build a
        view (i.e. weak reference) from it or a deep copy from it.
    :type copy: bool
    :param kwargs: Other parameters to be set

    """

    def __init__(
        self,
        scores: 'MessageMapContainer',
    ):
        self._pb_body = scores

    def __setitem__(
        self,
        key: str,
        value: Union[jina_pb2.NamedScoreProto, NamedScore, float, np.generic],
    ):
        if isinstance(value, jina_pb2.NamedScoreProto):
            self._pb_body[key].CopyFrom(value)
        elif isinstance(value, NamedScore):
            self._pb_body[key].CopyFrom(value._pb_body)
        elif isinstance(value, (float, int)):
            self._pb_body[key].value = value
        elif isinstance(value, np.generic):
            self._pb_body[key].value = value.item()
        else:
            raise TypeError(f'score is in unsupported type {typename(value)}')

    def __getitem__(
        self,
        key: str,
    ):
        return NamedScore(self._pb_body[key])

    def __delitem__(
        self,
        key: str,
    ):
        del self._pb_body[key]

    def __contains__(self, key: str):
        return key in self._pb_body

    def __iter__(self):
        for key in self._pb_body:
            yield key

    def __len__(self):
        return len(self._pb_body)



"""Argparser module for Flow"""
from .base import set_base_parser
from .helper import add_arg_group, KVAppendAction


def mixin_flow_features_parser(parser):
    """Add the arguments for the Flow features to the parser

    :param parser: the parser configure
    """
    from ..enums import FlowInspectType

    gp = add_arg_group(parser, title='Flow Feature')

    gp.add_argument('--uses', type=str, help='The YAML file represents a flow')
    gp.add_argument(
        '--env',
        action=KVAppendAction,
        metavar='KEY: VALUE',
        nargs='*',
        help='The map of environment variables that are available inside runtime',
    )

    gp.add_argument(
        '--inspect',
        type=FlowInspectType.from_string,
        choices=list(FlowInspectType),
        default=FlowInspectType.COLLECT,
        help='''
    The strategy on those inspect pods in the flow.

    If `REMOVE` is given then all inspect pods are removed when building the flow.
    ''',
    )


def set_flow_parser(parser=None, with_identity=False):
    """Set the parser for the flow

    :param parser: an (optional) initial parser to build upon
    :param with_identity: if to include identity in the parser
    :return: the parser
    """
    from .peapods.base import mixin_base_ppr_parser

    if not parser:
        parser = set_base_parser()

    mixin_base_ppr_parser(parser, with_identity=with_identity)

    parser.set_defaults(workspace='./')

    mixin_flow_features_parser(parser)

    return parser



import os

import grpc

from ..prefetch import PrefetchCaller
from ...zmq.asyncio import AsyncNewLoopRuntime
from ....zmq import AsyncZmqlet
from .....proto import jina_pb2_grpc

__all__ = ['GRPCRuntime']


class GRPCPrefetchCall(jina_pb2_grpc.JinaRPCServicer):
    """JinaRPCServicer """

    def __init__(self, args, zmqlet):
        super().__init__()
        self._servicer = PrefetchCaller(args, zmqlet)
        self.Call = self._servicer.send
        self.close = self._servicer.close


class GRPCRuntime(AsyncNewLoopRuntime):
    """Runtime for gRPC."""

    async def async_setup(self):
        """
        The async method to setup.

        Create the gRPC server and expose the port for communication.
        """
        if not self.args.proxy and os.name != 'nt':
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')

        self.server = grpc.aio.server(
            options=[
                ('grpc.max_send_message_length', -1),
                ('grpc.max_receive_message_length', -1),
            ]
        )
        self.zmqlet = AsyncZmqlet(self.args, logger=self.logger)
        self._prefetcher = GRPCPrefetchCall(self.args, self.zmqlet)
        jina_pb2_grpc.add_JinaRPCServicer_to_server(self._prefetcher, self.server)
        bind_addr = f'{self.args.host}:{self.args.port_expose}'
        self.server.add_insecure_port(bind_addr)
        await self.server.start()

    async def async_cancel(self):
        """The async method to stop server."""
        await self.server.stop(0)
        await self._prefetcher.close()

    async def async_run_forever(self):
        """The async running of server."""
        self.is_ready_event.set()
        await self.server.wait_for_termination()
        self.zmqlet.close()



from math import inf
from typing import Optional, Tuple, Union

import numpy as np

from jina import Document

if False:
    from .document import DocumentArray
    from .memmap import DocumentArrayMemmap


class DocumentArrayNeuralOpsMixin:
    """ A mixin that provides match functionality to DocumentArrays """

    def match(
        self,
        darray: Union['DocumentArray', 'DocumentArrayMemmap'],
        metric: str = 'cosine',
        limit: Optional[int] = inf,
        is_distance: bool = False,
    ) -> None:
        """Compute embedding based nearest neighbour in `another` for each Document in `self`,
        and store results in `matches`.

        Note:
            - If metric is 'cosine' it uses the cosine **distance**.
            - If metric is 'euclidean' it uses the euclidean **distance**.
            - If metric is 'euclidean_squared' it uses the euclidean **distance** squared.

        :param darray: the other DocumentArray or DocumentArrayMemmap to match against
        :param metric: the distance metric
        :param limit: the maximum number of matches, when not given
                      all Documents in `another` are considered as matches
        :param is_distance: Boolean flag informing if `metric` values want to be considered as distances or scores.
        """

        def invert_if_score(
            x: 'np.ndarray',
        ) -> 'np.ndarray':
            """Invert values to scores if `is_distance=False` according to the metric that is passed.

            :param x: input np.ndarray
            :return: A np.ndarray with distances if `is_distance=True`, scores if distance=False`.
            """
            if metric == 'cosine':
                if is_distance:
                    x = x
                else:
                    x = 1 - x
            elif metric == 'euclidean_squared' or metric == 'euclidean':
                if is_distance:
                    x = x
                else:
                    x = 1 / (x + 1)
            return x

        X = np.stack(self.get_attributes('embedding'))
        Y = np.stack(darray.get_attributes('embedding'))
        limit = min(limit, len(darray))

        dists = _compute_distances(X, Y, metric)
        idx, dist = self._get_sorted_smallest_k(dists, limit)
        dist = invert_if_score(dist)

        for _q, _ids, _dists in zip(self, idx, dist):
            for _id, _dist in zip(_ids, _dists):
                d = Document(darray[int(_id)], copy=True)
                d.scores[metric] = _dist
                _q.matches.append(d)

    @staticmethod
    def _get_sorted_smallest_k(
        dists: 'np.ndarray', top_k: int
    ) -> Tuple['np.ndarray', 'np.ndarray']:
        """Finds the smallest `top_k` values (and its indices) from `dists`. Returns both indices and values.

        :param dists: np.ndarray of distances
        :param top_k: number of values to retrieve
        :return: Lowest values in dists and the indices of those values
        """
        if top_k >= dists.shape[1]:
            indices = dists.argsort(axis=1)[:, :top_k]
            dists = np.take_along_axis(dists, indices, axis=1)
        else:
            indices_ps = dists.argpartition(kth=top_k, axis=1)[:, :top_k]
            dists = np.take_along_axis(dists, indices_ps, axis=1)
            indices_fs = dists.argsort(axis=1)
            indices = np.take_along_axis(indices_ps, indices_fs, axis=1)
            dists = np.take_along_axis(dists, indices_fs, axis=1)

        return indices, dists


def _compute_distances(x: 'np.ndarray', y: 'np.ndarray', metric: str) -> 'np.ndarray':
    """Computes the distance between each row of X and each row on Y according to `metric`.
    - Let `n_X = X.shape[0]`
    - Let `n_Y = Y.shape[0]`
    - Returns a matrix `dist` of shape `(n_X, n_Y)` with `dist[i,j] = metric(X[i], X[j])`.
    :param x: np.ndarray of ndim 2
    :param y:  np.ndarray of ndim 2
    :param metric: string describing the metric type
    :return: np.ndarray of ndim 2
    """
    if metric == 'cosine':
        dists = _cosine_distance(x, y)
    elif metric == 'euclidean_squared':
        dists = _euclidean_distance_squared(x, y)
    elif metric == 'euclidean':
        dists = np.sqrt(_euclidean_distance_squared(x, y))
    else:
        raise ValueError(f'Input metric={metric} not valid')
    return dists


def _cosine_distance(x: 'np.ndarray', y: 'np.ndarray') -> 'np.ndarray':
    """Cosine distance between each row in X and each row in Y.
    :param x: np.ndarray with ndim=2
    :param y: np.ndarray with ndim=2
    :return: np.ndarray  with ndim=2
    """
    return 1 - np.dot(x, y.T) / np.outer(
        np.linalg.norm(x, axis=1), np.linalg.norm(y, axis=1)
    )


def _euclidean_distance_squared(x: 'np.ndarray', y: 'np.ndarray') -> 'np.ndarray':
    """Euclidean (squared) distance between each row in X and each row in Y.
    :param x: np.ndarray with ndim=2
    :param y: np.ndarray with ndim=2
    :return: np.ndarray with ndim=2
    """
    return (
        np.sum(y ** 2, axis=1)
        + np.sum(x ** 2, axis=1)[:, np.newaxis]
        - 2 * np.dot(x, y.T)
    )



from pathlib import Path
from shutil import rmtree
from copy import deepcopy
from typing import Union

from jina.enums import RemoteWorkspaceState
from jina.helper import colored
from .base import BaseStore
from ..dockerize import Dockerizer
from ..models import DaemonID
from ..models.workspaces import (
    WorkspaceArguments,
    WorkspaceItem,
    WorkspaceMetadata,
    WorkspaceStoreStatus,
)


class WorkspaceStore(BaseStore):
    """A store of workspaces built by Daemon as docker objects"""

    _kind = 'workspace'
    _status_model = WorkspaceStoreStatus

    @BaseStore.dump
    def add(self, id: DaemonID, value: RemoteWorkspaceState, **kwargs):
        """Add a workspace to the store

        :param id: workspace id
        :param value: state of the workspace
        :param kwargs: keyword args
        :return: workspace id
        """
        if isinstance(value, RemoteWorkspaceState):
            self[id] = WorkspaceItem(state=value)
        return id

    @BaseStore.dump
    def update(
        self,
        id: DaemonID,
        value: Union[
            WorkspaceItem, RemoteWorkspaceState, WorkspaceArguments, WorkspaceMetadata
        ],
        **kwargs,
    ) -> DaemonID:
        """Update the workspace

        :param id: workspace id
        :param value: workspace value to be added
        :param kwargs: keyword args
        :raises KeyError: if id doesn't exist in the store
        :return: workspace id
        """
        if id not in self:
            raise KeyError(f'workspace {id} not found in store')

        if isinstance(value, WorkspaceItem):
            self[id] = value
        elif isinstance(value, WorkspaceArguments):
            self[id].arguments = value
        elif isinstance(value, WorkspaceMetadata):
            self[id].metadata = value
        elif isinstance(value, RemoteWorkspaceState):
            self[id].state = value
        else:
            self._logger.error(f'invalid arguments for workspace: {value}')
        return id

    def rm_files(self, id: DaemonID, logs: bool = False) -> None:
        """Remove files from workspace

        :param id: workspace id
        :param logs: True if log files should be removed, defaults to False
        """
        if self[id].metadata:
            workdir = self[id].metadata.workdir
            if not workdir or not Path(workdir).is_dir():
                self._logger.info(f'there\'s nothing to remove in workdir {workdir}')
                return
            if logs:
                self._logger.info(f'asked to remove complete directory: {workdir}')
                rmtree(workdir)
                self[id].metadata.workdir = ''
            else:
                for path in Path(workdir).rglob('[!logging.log]*'):
                    if path.is_file():
                        self._logger.debug(f'file to be deleted: {path}')
                        path.unlink()

    def rm_network(self, id: DaemonID) -> None:
        """Remove docker network

        :param id: workspace id
        """
        try:
            network = self[id].metadata.network
            # TODO: check what containers are using this network
            status = False
            if id in Dockerizer.networks:
                status = Dockerizer.rm_network(network)
            else:
                self._logger.info(f'no network to delete for id {colored(id, "cyan")}')
            if network and status:
                self[id].metadata.network = None
        except AttributeError as e:
            self._logger.info(f'there\'s no network to remove {e!r}')

    def rm_container(self, id: DaemonID) -> None:
        """Remove docker container

        :param id: workspace id
        """
        try:
            container_id = self[id].metadata.container_id
            if id in Dockerizer.containers:
                Dockerizer.rm_container(container_id)
                assert id not in Dockerizer.containers
                self._logger.success(
                    f'container {colored(container_id, "cyan")} is successfully removed'
                )
            else:
                self._logger.debug(
                    f'no container to delete for id {colored(id, "cyan")}'
                )
            if container_id:
                self[id].metadata.container_id = None
        except AttributeError as e:
            self._logger.error(f'there\'s no containers to remove {e!r}')
        except AssertionError as e:
            self._logger.error(f'something went wrong while removing the container')
            raise
        except Exception as e:
            self._logger.error(f'something went wrong while removing the container')
            raise

    @BaseStore.dump
    def delete(
        self,
        id: DaemonID,
        container: bool = True,
        network: bool = True,
        files: bool = True,
        everything: bool = False,
        **kwargs,
    ) -> None:
        """Delete a workspace from the store

        :param id: workspace id
        :param container: True if workspace container needs to be removed, defaults to True
        :param network: True if network needs to be removed, defaults to True
        :param files: True if files in the workspace needs to be removed, defaults to True
        :param everything: True if everything needs to be removed, defaults to False
        :param kwargs: keyword args
        :raises KeyError: if id doesn't exist in the store
        """
        if everything:
            container = True
            network = True
            files = True

        if id not in self:
            raise KeyError(f'{colored(str(id), "cyan")} not found in store.')

        # Peas/Pods/Flows need to be deleted before networks, files etc can be deleted
        if everything and self[id].metadata:
            from . import get_store_from_id

            ids_to_delete = list(self[id].metadata.managed_objects)
            for managed_object in ids_to_delete:
                get_store_from_id(managed_object).delete(id=managed_object)

        if container:
            self.rm_container(id)
        if network:
            self.rm_network(id)
        if files:
            # TODO: deleting files when not deleting container seems unwise
            self.rm_files(id, logs=everything)

        if everything:
            del self[id]
            self._logger.success(
                f'{colored(str(id), "cyan")} is released from the store.'
            )

    def clear(self, **kwargs) -> None:
        """Delete all the objects in the store

        :param kwargs: keyward args
        """

        _status = deepcopy(self.status)
        for k in _status.items.keys():
            self.delete(id=k, workspace=True, **kwargs)
